@@ -1,12 +1,12 @@
 # -*- coding: utf-8 -*-
 #
 #
-#  TheVirtualBrain-Scientific Package. This package holds all simulators, and 
+# TheVirtualBrain-Scientific Package. This package holds all simulators, and
 # analysers necessary to run brain-simulations. You can use it stand alone or
 # in conjunction with TheVirtualBrain-Framework Package. See content of the
 # documentation-folder for more details. See also http://www.thevirtualbrain.org
 #
-# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
+# (c) 2012-2022, Baycrest Centre for Geriatric Care ("Baycrest") and others
 #
 # This program is free software: you can redistribute it and/or modify it under the
 # terms of the GNU General Public License as published by the Free Software Foundation,
@@ -46,7 +46,7 @@ will be consistent with Monitor periods corresponding to any of [4096, 2048, 102
 """
 import abc
 import functools
-import numpy as np
+import numpy
 import scipy.integrate
 from . import noise
 from .common import get_logger, simple_gen_astr
@@ -98,13 +98,19 @@ class Integrator(HasTraits):
 
     clamped_state_variable_indices = NArray(
         dtype=int,
-        label="indices of the state variables to be clamped by the integrators to the values in the clamped_values array",
+        label="indices of the state variables to be clamped by the integrators "
+              "to the values in the clamped_values array",
         required=False)
 
     clamped_state_variable_values = NArray(
         label="The values of the state variables which are clamped ",
         required=False)
 
+    _bounded_integration_state_variable_indices = None
+    _integration_state_variable_boundaries = None
+    _clamped_integration_state_variable_indices = None
+    _clamped_integration_state_variable_values = None
+
     @abc.abstractmethod
     def scheme(self, X, dfun, coupling, local_coupling, stimulus):
         """
@@ -114,17 +120,16 @@ class Integrator(HasTraits):
 
         """
 
-    def bound_state(self, X):
-        for sv_ind, sv_bounds in \
-                zip(self.bounded_state_variable_indices,
-                    self.state_variable_boundaries):
-            if sv_bounds[0] is not None:
-                X[sv_ind][X[sv_ind] < sv_bounds[0]] = sv_bounds[0]
-            if sv_bounds[1] is not None:
-                X[sv_ind][X[sv_ind] > sv_bounds[1]] = sv_bounds[1]
+    def set_random_state(self, random_state):
+        self.log.warning("random_state supplied for non-stochastic integration")
 
-    def clamp_state(self, X):
-        X[self.clamped_state_variable_indices] = self.clamped_state_variable_values
+    def configure(self):
+        # Set default configurations:
+        self._clamped_integration_state_variable_indices = self.clamped_state_variable_indices
+        self._clamped_integration_state_variable_values = self.clamped_state_variable_values
+        self._bounded_integration_state_variable_indices = self.bounded_state_variable_indices
+        self._integration_state_variable_boundaries = self.state_variable_boundaries
+        super(Integrator, self).configure()
 
     def configure_boundaries(self, model):
         if model.state_variable_boundaries is not None:
@@ -133,15 +138,70 @@ class Integrator(HasTraits):
             for sv, sv_bounds in model.state_variable_boundaries.items():
                 indices.append(model.state_variables.index(sv))
                 boundaries.append(sv_bounds)
-            sort_inds = np.argsort(indices)
-            self.bounded_state_variable_indices = np.array(indices)[sort_inds]
-            self.state_variable_boundaries = np.array(boundaries).astype("float64")[sort_inds]
-        else:
-            self.bounded_state_variable_indices = None
-            self.state_variable_boundaries = None
+            sort_inds = numpy.argsort(indices)
+            self.bounded_state_variable_indices = numpy.array(indices)[sort_inds]
+            self.state_variable_boundaries = numpy.array(boundaries).astype("float64")[sort_inds]
+            self._bounded_integration_state_variable_indices = numpy.copy(self.bounded_state_variable_indices)
+            self._integration_state_variable_boundaries = numpy.copy(self.state_variable_boundaries)
 
-    def set_random_state(self, random_state):
-        self.log.warn("random_state supplied for non-stochastic integration")
+    def reconfigure_boundaries_and_clamping_for_integration_state_variables(self, model):
+        integration_state_variable_indices = numpy.where(model.state_variable_mask)[0].tolist()
+        if self.state_variable_boundaries is not None:
+            # If there are any state_variable_boundaries...
+            bounded_integration_state_variable_indices = []
+            integration_state_variable_boundaries = []
+            # ...for each one of the bounded state variable indices and boundary values...
+            for bound_sv_ind, bounds in zip(self._bounded_integration_state_variable_indices,
+                                            self.state_variable_boundaries):
+                # ...if the boundary indice corresponds to an integrated state variable...
+                if bound_sv_ind in integration_state_variable_indices:
+                    # ...add its integration state vector indice...
+                    bounded_integration_state_variable_indices.append(
+                        integration_state_variable_indices.index(bound_sv_ind))
+                    # ...and the corresponding boundaries
+                    integration_state_variable_boundaries.append(bounds)
+            self._bounded_integration_state_variable_indices = \
+                numpy.array(bounded_integration_state_variable_indices)
+            self._integration_state_variable_boundaries = \
+                numpy.array(integration_state_variable_boundaries)
+        if self.clamped_state_variable_values is not None:
+            # If there are any clamped values...
+            clamped_integration_state_variable_indices = []
+            clamped_integration_state_variable_values = []
+            # ...for each one of the clamped state variable indices and clamped values...
+            for clamp_sv_ind, clampval in zip(self.clamped_state_variable_indices,
+                                              self.clamped_state_variable_values):
+                # ...if the clamped indice corresponds to an integrated state variable...
+                if clamp_sv_ind in integration_state_variable_indices:
+                    # ...add its integration state vector indice...
+                    clamped_integration_state_variable_indices.append(
+                        integration_state_variable_indices.index(clamp_sv_ind))
+                    # ...and the corresponding clamped value
+                    clamped_integration_state_variable_values.append(clampval)
+            self._clamped_integration_state_variable_indices = \
+                numpy.array(clamped_integration_state_variable_indices)
+            self._clamped_integration_state_variable_values = \
+                numpy.array(clamped_integration_state_variable_values)
+
+    def _bound_state(self, X, indices, boundaries):
+        for sv_ind, sv_bounds in zip(indices, boundaries):
+            if sv_bounds[0] is not None:
+                X[sv_ind][X[sv_ind] < sv_bounds[0]] = sv_bounds[0]
+            if sv_bounds[1] is not None:
+                X[sv_ind][X[sv_ind] > sv_bounds[1]] = sv_bounds[1]
+
+    def bound_state(self, X):
+        self._bound_state(X, self.bounded_state_variable_indices, self.state_variable_boundaries)
+
+    def bound_integration_state(self, X):
+        self._bound_state(X, self._bounded_integration_state_variable_indices,
+                          self._integration_state_variable_boundaries)
+
+    def clamp_state(self, X):
+        X[self.clamped_state_variable_indices] = self.clamped_state_variable_values
+
+    def clamp_integration_state(self, X):
+        X[self._clamped_integration_state_variable_indices] = self._clamped_integration_state_variable_values
 
     def bound_and_clamp(self, state):
         # If there is a state boundary...
@@ -153,6 +213,16 @@ class Integrator(HasTraits):
             # ...use the integrator's clamp_state
             self.clamp_state(state)
 
+    def integration_bound_and_clamp(self, state):
+        # If there is a state boundary...
+        if self._integration_state_variable_boundaries is not None:
+            # ...use the integrator's bound_state
+            self.bound_integration_state(state)
+        # If there is a state clamping...
+        if self._clamped_integration_state_variable_values is not None:
+            # ...use the integrator's clamp_state
+            self.clamp_integration_state(state)
+
     def integrate_with_update(self, X, model, coupling, local_coupling, stimulus):
         temp = model.update_state_variables_before_integration(X, coupling, local_coupling, stimulus)
         if temp is not None:
@@ -166,8 +236,8 @@ class Integrator(HasTraits):
         return X
 
     def integrate(self, X, model, coupling, local_coupling, stimulus):
-        X[model.state_variables_mask] = self.scheme(X, model.dfun, coupling, local_coupling, stimulus)
-        self.bound_and_clamp(X)
+        X[model.state_variable_mask] = self.scheme(X[model.state_variable_mask],
+                                                   model.dfun, coupling, local_coupling, stimulus)
         return X
 
     def __str__(self):
@@ -227,6 +297,7 @@ class HeunDeterministic(Integrator):
     """
 
     _ui_name = "Heun"
+    n_dx = 2
 
     def scheme(self, X, dfun, coupling, local_coupling, stimulus):
         r"""
@@ -243,18 +314,13 @@ class HeunDeterministic(Integrator):
         #import pdb; pdb.set_trace()
         m_dx_tn = dfun(X, coupling, local_coupling)
         inter = X + self.dt * (m_dx_tn + stimulus)
-        if self.state_variable_boundaries is not None:
-            self.bound_state(inter)
-        if self.clamped_state_variable_values is not None:
-            self.clamp_state(inter)
+        self.integration_bound_and_clamp(inter)
 
         dX = (m_dx_tn + dfun(inter, coupling, local_coupling)) * self.dt / 2.0
 
         X_next = X + dX + self.dt * stimulus
-        if self.state_variable_boundaries is not None:
-            self.bound_state(X_next)
-        if self.clamped_state_variable_values is not None:
-            self.clamp_state(X_next)
+        self.integration_bound_and_clamp(X_next)
+
         return X_next
 
 
@@ -266,6 +332,7 @@ class HeunStochastic(IntegratorStochastic):
     """
 
     _ui_name = "Stochastic Heun"
+    n_dx = 2
 
     def scheme(self, X, dfun, coupling, local_coupling, stimulus):
         """
@@ -292,18 +359,12 @@ class HeunStochastic(IntegratorStochastic):
         noise *= noise_gfun
 
         inter = X + self.dt * m_dx_tn + noise + self.dt * stimulus
-        if self.state_variable_boundaries is not None:
-            self.bound_state(inter)
-        if self.clamped_state_variable_values is not None:
-            self.clamp_state(inter)
+        self.integration_bound_and_clamp(inter)
 
         dX = (m_dx_tn + dfun(inter, coupling, local_coupling)) * self.dt / 2.0
 
         X_next = X + dX + noise + self.dt * stimulus
-        if self.state_variable_boundaries is not None:
-            self.bound_state(X_next)
-        if self.clamped_state_variable_values is not None:
-            self.clamp_state(X_next)
+        self.integration_bound_and_clamp(X_next)
 
         return X_next
 
@@ -318,6 +379,7 @@ class EulerDeterministic(Integrator):
     """
 
     _ui_name = "Euler"
+    n_dx = 1
 
     def scheme(self, X, dfun, coupling, local_coupling, stimulus):
         r"""
@@ -332,10 +394,8 @@ class EulerDeterministic(Integrator):
         self.dX = dfun(X, coupling, local_coupling) 
 
         X_next = X + self.dt * (self.dX + stimulus)
-        if self.state_variable_boundaries is not None:
-            self.bound_state(X_next)
-        if self.clamped_state_variable_values is not None:
-            self.clamp_state(X_next)
+        self.integration_bound_and_clamp(X_next)
+
         return X_next
 
 
@@ -349,6 +409,7 @@ class EulerStochastic(IntegratorStochastic):
     """
 
     _ui_name = "Euler-Maruyama"
+    n_dx = 1
 
     def scheme(self, X, dfun, coupling, local_coupling, stimulus):
         r"""
@@ -369,10 +430,8 @@ class EulerStochastic(IntegratorStochastic):
         dX = dfun(X, coupling, local_coupling) * self.dt 
         noise_gfun = self.noise.gfun(X)
         X_next = X + dX + noise_gfun * noise + self.dt * stimulus
-        if self.state_variable_boundaries is not None:
-            self.bound_state(X_next)
-        if self.clamped_state_variable_values is not None:
-            self.clamp_state(X_next)
+        self.integration_bound_and_clamp(X_next)
+
         return X_next
 
 
@@ -383,6 +442,7 @@ class RungeKutta4thOrderDeterministic(Integrator):
     """
 
     _ui_name = "Runge-Kutta 4th order"
+    n_dx = 4
 
     def scheme(self, X, dfun, coupling, local_coupling=0.0, stimulus=0.0):
         r"""
@@ -410,31 +470,23 @@ class RungeKutta4thOrderDeterministic(Integrator):
 
         k1 = dfun(X, coupling, local_coupling)
         inter_k1 = X + dt2 * k1
-        if self.state_variable_boundaries is not None:
-            self.bound_state(inter_k1)
-        if self.clamped_state_variable_values is not None:
-            self.clamp_state(inter_k1)
+        self.integration_bound_and_clamp(inter_k1)
+
         k2 = dfun(inter_k1, coupling, local_coupling)
         inter_k2 = X + dt2 * k2
-        if self.state_variable_boundaries is not None:
-            self.bound_state(inter_k2)
-        if self.clamped_state_variable_values is not None:
-            self.clamp_state(inter_k2)
+        self.integration_bound_and_clamp(inter_k2)
+
         k3 = dfun(inter_k2, coupling, local_coupling)
         inter_k3 = X + dt * k3
-        if self.state_variable_boundaries is not None:
-            self.bound_state(inter_k3)
-        if self.clamped_state_variable_values is not None:
-            self.clamp_state(inter_k3)
+        self.integration_bound_and_clamp(inter_k3)
+
         k4 = dfun(inter_k3, coupling, local_coupling)
 
         dX = dt6 * (k1 + 2.0 * k2 + 2.0 * k3 + k4)
 
         X_next = X + dX + self.dt * stimulus
-        if self.state_variable_boundaries is not None:
-            self.bound_state(X_next)
-        if self.clamped_state_variable_values is not None:
-            self.clamp_state(X_next)
+        self.integration_bound_and_clamp(X_next)
+
         return X_next
 
 
@@ -450,6 +502,7 @@ class Identity(Integrator):
     """
 
     _ui_name = "Difference equation"
+    n_dx = 1
 
     def scheme(self, X, dfun, coupling=None, local_coupling=0.0, stimulus=0.0):
         """
@@ -461,7 +514,33 @@ class Identity(Integrator):
 
         """
 
-        return dfun(X, coupling, local_coupling) + stimulus
+        X_next = dfun(X, coupling, local_coupling) + stimulus
+        self.integration_bound_and_clamp(X_next)
+        return X_next
+
+
+class IdentityStochastic(IntegratorStochastic):
+    """
+    A stochastic variant of the Identity integrator.  Together
+    with time delays, this allows for MVAR models. 
+    """
+
+    _ui_name = "Stochastic difference equation (MVAR)"
+    n_dx = 1
+
+    def scheme(self, X, dfun, coupling=None, local_coupling=0.0, stimulus=0.0):
+        """
+        The stochastic identity scheme simply returns the results of the dfun and
+        the gfun and stimulus.
+
+        .. math::
+            x_{n+1} = f(x_{n}) + g(X_n) Z_1
+
+        """
+        z = self.noise.generate(X.shape) * self.noise.gfun(X)
+        X_next = dfun(X, coupling, local_coupling) + z + stimulus
+        self.integration_bound_and_clamp(X_next)
+        return X_next
 
 
 class SciPyODEBase(object):
@@ -501,10 +580,7 @@ class SciPyODE(SciPyODEBase):
 
     def scheme(self, X, dfun, coupling, local_coupling, stimulus):
         X_next = self._apply_ode(X, dfun, coupling, local_coupling, stimulus)
-        if self.state_variable_boundaries is not None:
-            self.bound_state(X_next)
-        if self.clamped_state_variable_values is not None:
-            self.clamp_state(X_next)
+        self.integration_bound_and_clamp(X_next)
         return X_next
 
 class SciPySDE(SciPyODEBase):
@@ -512,10 +588,7 @@ class SciPySDE(SciPyODEBase):
     def scheme(self, X, dfun, coupling, local_coupling, stimulus):
         X_next = self._apply_ode(X, dfun, coupling, local_coupling, stimulus)
         X_next += self.noise.gfun(X) * self.noise.generate(X.shape)
-        if self.state_variable_boundaries is not None:
-            self.bound_state(X_next)
-        if self.clamped_state_variable_values is not None:
-            self.clamp_state(X_next)
+        self.integration_bound_and_clamp(X_next)
         return X_next
 
 
